/*
 * @file cmd_flboot.c
 * @brief solar L8 board boot flow command.
 * @author m.hirobe@co-nss.co.jp
 * @date 2019-08-30
 */

#include <asm/arch/imx-regs.h>
#include <asm/arch/crm_regs.h>
#include <asm/arch/mx6-pins.h>
#include <asm/arch/sys_proto.h>
#include <asm/gpio.h>
#include <watchdog.h>
#include <configs/mx6ull_solar_l8.h>
#include <mx6ull_solar_l8/fw_info.h>
#include <mx6ull_solar_l8/eeprom.h>

extern void hw_watchdog_init(void);

/*
 * @fn fw_setenv_macaddr
 * @brief set macaddress to enviroment variables.
 * @param u8 *ethaddr : mac address
 * @aram char *envname : enviroment variable name.
 * @return none
 */
static void fw_setenv_macaddr(u8 *ethaddr, char *envname)
{
	char macaddr_str[18];

	sprintf (macaddr_str, "%02X:%02X:%02X:%02X:%02X:%02X", ethaddr[0], ethaddr[1], ethaddr[2],
			ethaddr[3], ethaddr[4], ethaddr[5]);

	if (strcmp(envname, "ethaddr") && strcmp(envname, "wlanaddr")) {
		printf ("Enviroment name error : %s\n", envname);
	}
	else {
		setenv(envname, macaddr_str);
		/* mac address is auto write to ENET register from u-boot env callback function. */
		/* see on_ethaddr() in ./u-boot/net/eth-uclass.c */
	}
}

/*
 * @fn fw_setenv_rootfs
 * @brief set rootfs load enviroment variable.
 * @param : int bootsts : boot status.
 * @return : none
 */
#define MOUNTED_OPTION	"ro"
static void fw_setenv_rootfs(int bootsts)
{
	char rootfs_args[64];
	char rootfs_part[32];

	if ((bootsts & START_SIDE_MASK) == START_SIDE_A) {
		/* side A(normal) */
		sprintf (rootfs_part, "root=ubi0:rootfs_a");
	}
	else {
		/* side B(backup) */
		sprintf (rootfs_part, "root=ubi0:rootfs_b");
	}
	sprintf (rootfs_args, "%s ubi.mtd=8,4096 rootfstype=ubifs %s", rootfs_part, MOUNTED_OPTION);
	setenv("rootfs_args", rootfs_args);
}

/*
 * @fn fw_setenv_loadaddr
 * @brief set loadaddr enviroment variable.
 * @param : u32 loadaddr : loadaddr
 * @return : none
 */
static void fw_setenv_loadaddr(u32 loadaddr)
{
	char loadaddr_str[16];

	sprintf(loadaddr_str, "0x%08x", loadaddr);
	setenv("loadaddr", loadaddr_str);
}

/*
 * @fn fw_setenv_bootargs
 * @brief set bootargs enviroment variables.
 * @param : none
 * @return : none
 */
static void fw_setenv_bootargs(void)
{
	char flbootargs[128];
	char *rootfs_args;
	char *bootmode_args;
	char *debug_args;
	char ipaddr[] = "";

	rootfs_args = getenv("rootfs_args");
	bootmode_args = getenv("bootmode_args");
	debug_args = getenv("debug_args");

	if ((rootfs_args != NULL) && (bootmode_args != NULL)) {
		if (debug_args) {
			sprintf (flbootargs, "console=ttymxc0,115200 %s bootcmd=%s %s %s",
					rootfs_args, bootmode_args, ipaddr, debug_args);
		}
		else {
			sprintf (flbootargs, "console=ttymxc0,115200 %s bootcmd=%s %s",
					rootfs_args, bootmode_args, ipaddr);
		}

		if (strcmp(bootmode_args, "test") == 0) {
			printf("+++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
			printf(" Test mode start ...                                   \n");
			printf("  - bootargs : %s\n", flbootargs);
			printf("+++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
		}

		setenv("bootargs", flbootargs);
	}
}

/*
 * @fn fw_kernel_load
 * @brief kernel image load.
 * @param : u32 loadaddr : kernel image load dst address.
 * @param : u32 imageaddr : kernel image load src address.
 * @return : 0 = success / 1 = error
 */
static int fw_kernel_load(u32 loadaddr, u32 imageaddr)
{
	char nand_cmd[128];

	sprintf (nand_cmd, "nand read 0x%x 0x%x 0x800000", loadaddr, imageaddr);

	/* when nand command success, return 0. */
	return run_command(nand_cmd, 0);
}

/*
 * @fn fw_version_display
 * @brief fw version display.
 * @param : u32 imageaddr : kenrel image address.
 * @param : u8 *version : fw version array top address.
 * @param : u8 *build_ver : fw build version array top address.
 * @return : none
 */
static void fw_version_display(u32 imageaddr, u8 *version, u8 *build_ver)
{
	char version_str[8], buildver_str[8];

	if (version) {
		sprintf (version_str, "%c%c.%c%c", version[0], version[1], version[2], version[3]);
	}
	else {
		sprintf (version_str, "NONE");
	}

	if (build_ver) {
		strncpy(buildver_str, (char *)build_ver, 4);
		buildver_str[4] = '\0';
	}
	else {
		sprintf (buildver_str, "NONE");
	}
	printf("+++ Barner +++++++++++++++++++++++++++++++++++\n");
	printf(" Kernel Image Address : %#x\n", imageaddr);
	printf("              Version : %s\n", version_str);
	printf("        Build Version : %s\n", buildver_str);
	printf("++++++++++++++++++++++++++++++++++++++++++++++\n");
}

/*
 * @fn do_flboot
 * @brief NAND Flashboot command.
 * @param cmd_tbl_t *cmdtp : command table pointer.
 * @param int flag : command flag.
 * @param int argc : arg strings num.
 * @param char *argv[] : arg strings.
 * @return 1 only.
 */
#define WDOG_TIMEOUT	0xFF
static int do_flboot(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	u16 bootsts;
	u32 loadaddr = CONFIG_SYS_LOAD_ADDR;
	u32 imageaddr;
	struct boot_fw_eeprom *fw_info;
	char *env;

	fw_info = fw_eeprom_read();

	if ((env = getenv("bootflags")) != NULL) {
		fw_info->bootflags = (int)simple_strtol(env, NULL, 16);
	}
	bootsts = fw_info->bootflags;

	if ((bootsts & START_STS_MASK) == START_FLAG_BACKUP) {
		printf ("Backup start error...\n");
		return 1;
	}

	if (fw_eeprom_bootsts_update(fw_info)) {
		printf ("Boot status update error\n");
	}

	bootsts = fw_info->bootflags;

	switch (bootsts & START_SIDE_MASK) {
		case START_SIDE_A:
			imageaddr = fw_info->side_a.kernel_addr;
			fw_version_display(fw_info->side_a.kernel_addr, (u8 *)fw_info->side_a.version,
								(u8 *)fw_info->side_a.build_ver);
			break;
		case START_SIDE_B:
			imageaddr = fw_info->side_b.kernel_addr;
			fw_version_display(fw_info->side_b.kernel_addr, (u8 *)fw_info->side_b.version,
								(u8 *)fw_info->side_b.build_ver);
			break;
		default:
			imageaddr = 0x00200000;
			fw_version_display(imageaddr, NULL, NULL);
			break;
	}

	debug ("kernel image - nand offset : %#08x\n", imageaddr);

	fw_setenv_macaddr(fw_info->ethaddr, "ethaddr");
	fw_setenv_loadaddr(loadaddr);
	fw_setenv_rootfs(bootsts);
	fw_setenv_bootargs();

	debug ("env get bootargs - %s\n", getenv("bootargs"));

	if (fw_kernel_load(loadaddr, imageaddr) != 0) {
		printf ("Kernel Nand Read Error...\n");
		return 1;
	}

	if ((bootsts & START_STS_MASK) != START_FLAG_OK) {
		printf ("Watchdog Timer start...\n");

		hw_watchdog_init();
	}

	if (run_command("run nandboot", 0) != 0) {
		printf ("kernel load error...\n");
	}

	return 1;
}

U_BOOT_CMD(
	flboot, 1, 1, do_flboot,
	"Memory image BOOT (main BOOT command)",
	""
);
