/*
 * @file eeprom.c
 * @brief eeprom access apis for mx6ull solar l8 board.
 * @author m.hirobe@co-nss.co.jp
 * @date 2019-08-29
 */
#include <asm/arch/iomux.h>
#include <asm/arch/imx-regs.h>
#include <asm/arch/mx6-pins.h>
#include <asm/arch/sys_proto.h>
#include <asm/gpio.h>
#include <asm/imx-common/iomux-v3.h>
#include <asm/imx-common/boot_mode.h>
#include <asm/imx-common/mxc_i2c.h>
#include <asm/io.h>
#include <common.h>
#include <dm.h>
#include <i2c.h>
#include <configs/mx6ull_solar_l8.h>
#include <mx6ull_solar_l8/fw_info.h>


static struct boot_fw_eeprom eeprom_fw_info;
static struct udevice *bus;
static struct udevice *dev;

/*
 * @fn __eeprom_off_conv
 * @brief eeprom bank num and data offset convert to eeprom offset.
 * @param : char *top : data offset top address.
 * @param : char *ptr : data offset address.
 * @param : int b : eeprom bank num.
 * @return : eeprom offset.
 */
static inline off_t __eeprom_off_conv(uint8_t *top, uint8_t *ptr, int b)
{
	return (off_t)(((uintptr_t)ptr - (uintptr_t)top) + (b * sizeof(struct boot_fw_eeprom)));
}

static int __eeprom_init(void)
{
	if (uclass_get_device_by_seq(UCLASS_I2C, I2C_EEPROM_BUSNO, &bus) != 0) {
		printf ("No bus %d\n", I2C_EEPROM_BUSNO);
		return -1;
	}

	if (dm_i2c_probe(bus, I2C_EEPROM_SLAVE_ADDR, 0, &dev) != 0) {
		printf ("Can't find device id = %#x, on bus %d\n", I2C_EEPROM_SLAVE_ADDR, I2C_EEPROM_BUSNO);
		return -1;
	}

	return 0;
}
/*
 * @fn __eeprom_read
 * @brief read from I2C_EEPROM_SLAVE_ADDR eeprom device.
 * @param unsigned int offset : read start offset.
 * @param uint8_t *buffer : read buffer address.
 * @param unsigned int size : read size.
 * @return 0 = success / not 0 = error
 */
static int __eeprom_read(unsigned int offset, uint8_t *buffer, int size)
{
	return dm_i2c_read(dev, offset, buffer, size);
}

/*
 * @fn __eeprom_write
 * @brief write to I2C_EEPROM_SLAVE_ADDR eeprom device.
 * @param unsigned int offset : write start offset.
 * @param uint8_t *buffer : write buffer address.
 * @param unsigned int size : write size.
 * @return 0 = success / not 0 = error
 */
#define PAGESIZE	(16)
static int __eeprom_write(unsigned int offset, uint8_t *buffer, unsigned int size)
{
	int ret = 0, idx = 0, remain, writesize;
	struct dm_i2c_chip *i2c_chip = dev_get_parent_platdata(dev);

	if (!i2c_chip) {
		printf ("dev_get_parent_platdata error\n");
		return -1;
	}

	/* write protect control */
	i2c_chip->flags &= ~DM_I2C_CHIP_WR_ADDRESS;
	gpio_set_value(IMX_GPIO_NR(1,24), 0);
	udelay(100);
	remain = size;
	while (remain > 0) {
		if (remain >= PAGESIZE) {
			writesize = PAGESIZE;
		}
		else {
			writesize = remain;
		}
		ret = dm_i2c_write(dev, offset+idx, &buffer[idx], writesize);
		if (ret < 0) {
			break;
		}
		remain -= writesize;
		idx += writesize;
		udelay(6 * 1000);
	}
	gpio_set_value(IMX_GPIO_NR(1,24), 1);

	return ret;
}

/*
 * @fn fw_eeprom_init
 * @brief eeprom initialize
 * @param none
 * @return none
 */
void fw_eeprom_init(void)
{
	int bank, ret;
	struct boot_fw_eeprom eeprom_info[EEPROM_BANK_MAX] = {0x00};
	struct boot_fw_eeprom *fw_info;
	u32 eeplen = sizeof(*fw_info);

	if (__eeprom_init() != 0) {
		printf ("eeprom init failed\n");
		return;
	}

	for (bank = 0; bank < EEPROM_BANK_MAX; bank++) {
		fw_info = &eeprom_info[bank];
		if (__eeprom_read((bank * eeplen), (uint8_t *)fw_info, eeplen) != 0) {
			printf ("EEPROM Bank %d: Read Error(%#x)\n", bank, (bank * eeplen));
			memset (fw_info, 0x00, eeplen);
		}
	}

	for (bank = 0; bank < EEPROM_BANK_MAX; bank++) {
		fw_info = &eeprom_info[bank];

		printf ("Bank %d - EEPROM Magic Check...", bank);

		if ((fw_info->magic & 0xFFFFFFF00) != (EEPROM_MAGIC_JHRWL8 & 0xFFFFFFF00)) {
			printf ("Bad Magic Code (%#x)\n", fw_info->magic);
			continue;
		}

		printf ("OK (%#x)\n", fw_info->magic);

		printf ("Bank %d - EEPROM CRC Verify...", bank);
		if (crc32(0, (uchar*)fw_info, (eeplen - sizeof(fw_info->crc))) != fw_info->crc) {
			printf ("Bad CRC(%#x)\n", fw_info->crc);
			continue;
		}
		printf ("OK\n");
		break;
	}

	if (bank == EEPROM_BANK_0) {
		if (memcmp(fw_info, &eeprom_info[EEPROM_BANK_1], eeplen) != 0) {
			if (__eeprom_write(eeplen, (uint8_t *)fw_info, eeplen) != 0) {
				printf ("EERPOM Bank 1 : write error\n");
			}
		}
	}
	else if (bank == EEPROM_BANK_1) {
		if (__eeprom_write(0x00, (uint8_t *)fw_info, eeplen) != 0) {
			printf ("EEPROM Bank 0 : write error\n");
		}
	}
	else {
		printf ("No EEPROM Data. Set Default Data\n");
		for (bank = 0; bank < EEPROM_BANK_MAX; bank++) {
			fw_info = &eeprom_info[bank];
			memset (fw_info, 0xFF, eeplen);

			fw_info->magic = EEPROM_MAGIC_JHRWL8;
			eth_parse_enetaddr(__stringify(CONFIG_ETHADDR), fw_info->ethaddr);
			memset (fw_info->wlanaddr, 0x00, sizeof(fw_info->wlanaddr));
			fw_info->bootflags = (START_SIDE_A | START_FLAG_UPDATE);
			fw_info->side_a.kernel_addr = KERNEL_A_SIDE;
			memcpy(fw_info->side_a.version, FW_DEFAULT_VERSION, 4);
			memcpy(fw_info->side_a.build_ver, FW_DEFAULT_BUILD, 4);
			fw_info->side_b.kernel_addr = KERNEL_B_SIDE;
			memcpy(fw_info->side_b.version, FW_DEFAULT_VERSION, 4);
			memcpy(fw_info->side_b.build_ver, FW_DEFAULT_BUILD, 4);
			fw_info->crc = crc32(0, (uchar *)fw_info, (sizeof(*fw_info) - sizeof(fw_info->crc)));

			if ((ret = __eeprom_write((bank * eeplen), (uint8_t *)fw_info, eeplen)) != 0) {
				printf ("EEPROM Bank %d: Write error(%#x) ret:%d\n", bank, (bank * eeplen), ret);
			}
			else {
				printf ("EEPROM Bank %d: Write finish(%#x)\n", bank, (bank * eeplen));
			}
		}
	}

	memcpy(&eeprom_fw_info, fw_info, eeplen);

}

/*
 * @fn fw_eeprom_read
 * @brief read eeprom info.
 * @param : none
 * @return : eeprom info pointer.
 */
struct boot_fw_eeprom* fw_eeprom_read(void)
{
	return &eeprom_fw_info;
}

/*
 * @fn fw_eeprom_macaddr_read
 * @brief read macaddress from eeprom
 * @param none
 * @return pointer to mac address string.
 */
char *fw_eeprom_macaddr_read(void)
{
	u8 zero_mac[6] = {0};
	u8 ff_mac[6];
	u8 def_mac[6];

	memset (ff_mac, 0xff, sizeof(ff_mac));
	eth_parse_enetaddr(__stringify(CONFIG_ETHADDR), def_mac);

	if ((memcmp(eeprom_fw_info.ethaddr, zero_mac, sizeof(zero_mac)) == 0) ||
		(memcmp(eeprom_fw_info.ethaddr, ff_mac, sizeof(ff_mac)) == 0) ||
		(memcmp(eeprom_fw_info.ethaddr, def_mac, sizeof(def_mac)) == 0)) {
		return NULL;
	}

	return (char *)eeprom_fw_info.ethaddr;
}

/*
 * @fn fw_eeprom_bootsts_update
 * @brief boot side flag and boot status flas update.
 * @param : struct boot_fw_eeprom *fw_info : EEPROM FW info.
 * @return : 0 = success / not 0 = error
 */
int fw_eeprom_bootsts_update(struct boot_fw_eeprom *fw_info)
{
	int bank;
	u16 update = 0, now = 0;

	if (!fw_info) {
		printf ("boot status update invalid parameter\n");
		return 1;
	}

	now = fw_info->bootflags;

	switch (now) {
		case (START_SIDE_A | START_FLAG_OK):		/* side a, normal : 0x0a00 */
		case (START_SIDE_B | START_FLAG_OK):		/* side b, normal : 0x0b00 */
			update = 0;
			break;
		case (START_SIDE_A | START_FLAG_UPDATE):	/* side a, update : 0x0a01 */
			update = (START_SIDE_A | START_FLAG_FIRST);		/* side a, first boot : 0x0a02 */
			break;
		case (START_SIDE_B | START_FLAG_UPDATE):	/* side b, update : 0x0b01 */
			update = (START_SIDE_B | START_FLAG_FIRST);		/* side b, first boot : 0x0b02 */
			break;
		case (START_SIDE_A | START_FLAG_FIRST):		/* side a, first boot : 0x0a02 */
			update = (START_SIDE_B | START_FLAG_BACKUP);	/* side b, backup : 0x0b03 */
			break;
		case (START_SIDE_B | START_FLAG_FIRST):		/* side b, first boot : 0x0b02 */
			update = (START_SIDE_A | START_FLAG_BACKUP);	/* side a, backup : 0x0a03 */
			break;
		default:									/* other */
			update = (START_SIDE_A | START_FLAG_FIRST);		/* side a, first boot : 0x0a02 */
			break;
	}

	if (update) {
		fw_info->bootflags = update;
		fw_info->crc = crc32(0, (uchar *)fw_info, (sizeof(*fw_info) - sizeof(fw_info->crc)));

		for (bank = 0; bank < EEPROM_BANK_MAX; bank++) {
			printf ("Bank %d : Bootflags update (%#x -> %#x)...", bank, now, update);
			if (__eeprom_write(__eeprom_off_conv((uint8_t *)fw_info, (uint8_t *)&fw_info->bootflags, bank),
							(uint8_t *)&fw_info->bootflags, sizeof(fw_info->bootflags)) != 0) {
				printf ("NG\n");
				return 1;
			}
			printf ("OK\n");

			printf ("Bank %d : CRC update...\n", bank);
			if (__eeprom_write(__eeprom_off_conv((uint8_t *)fw_info, (uint8_t *)&fw_info->crc, bank),
							(uint8_t *)&fw_info->crc, sizeof(fw_info->crc)) != 0) {
				printf ("NG\n");
				return 1;
			}
			printf ("OK\n");
		}
	}

	return 0;
}
